/*
 *  Assembler code generator:
 */

#include "asm.h"

/*
 *  Generator error handling:
 */
static void gen_error(Program p, char *msg, ...)
{
	va_list args;

	va_start(args, msg);
	if (++(p->error_count) <= MAX_ERRORS)
		if (msg) {
			fprintf(stderr, "error: ");
			vfprintf(stderr, msg, args);
		}
	va_end(args);
}

/*
 *  Expression evaluator:
 */
long eval_expr(Program p, Expr e)
{
	long left, right;

	if (e == NULL)
		return 0;
	if (e->token) {
		switch (e->token->kind) {
		 case IDENT:
			gen_error(p, "Undefined label \"%s\" used in expression\n",
				e->token->name);
			break;
		 case INTVAL:
			return e->token->num;
		}
	}
	else {
		if (e->left && e->right) {
			left = eval_expr(p, e->left);
			right = eval_expr(p, e->right);
			switch (e->op) {
			 case LSHIFT:	return left << right;
			 case RSHIFT:	return left >> right;
			 case STAR:	return left * right;
			 case SLASH:	return left / right;
			 case MOD:	return left % right;
			 case PLUS:	return left + right;
			 case MINUS:	return left - right;
			 case BITWISE_AND: return left & right;
			 case BITWISE_OR:  return left | right;
			 case BITWISE_XOR: return left ^ right;
			}
		}
		else if (e->left)
			return eval_expr(p, e->left);
		else if (e->right) {
			switch (e->op) {
			 case PLUS:
				return eval_expr(p, e->right);
			 case MINUS:
				return -eval_expr(p, e->right);
			 case BITWISE_NOT:
				return ~eval_expr(p, e->right);
			}
		}
	}
	return 0;
}

int is_relative_branch(Expr e)
{
	if (e && (! e->token) && (! e->left) && (e->right)) {
		if ((e->op == PLUS) || (e->op == MINUS))
			return 1;
	}
	return 0;
}

/*
 *  Encode one instruction as hex:
 *  Return 0 on failure, 1 on success.
 *  Failure could happen if we cannot find that combination of
 *  operand modes, or if the opcode is wrong.
 *  This is the opposite of the virtual machine's decode function.
 *  It's a simple algorithm: walk through the hash table of
 *  operations and find the first operand mode-set which allows
 *  the instruction to be correctly encoded.
 */
static
int encode(Program p, long opcode, int numops,
	long *opmode, long *opvalue, int relative_branch, long *code)
{
	int i, n, ch;
	long hexcode;
	long offset;
	ValidMode *modetable;
	ValidMode *set;

	modetable = opvalid[opcode];
	if (modetable == NULL) {
		gen_error(p, "Invalid opcode %02.2lX in instruction %ld: ",
			opcode, p->addr);
		return 0;
	}

	(*code) = (opcode << 24);

	for (set=modetable; set->opchar[0] != '\0'; set++) {
		/* for all valid operand mode sets in this opcode: */
		if (numops != set->opcount)
			continue;
		hexcode = 0; /* reset operand encoding */
		for (i=0, n=0; ; i++) {
			/* check each mode, starting at src1 */
			if (i == 3) {
				(*code) |= hexcode;
				return 1; /* success */
			}
			ch = set->opchar[i];
			if (ch == 'A') {
				if (relative_branch) {
					if ((set+1)->opchar[0] == 'S') {
						/* use S rel mode instead */
						break;
					}
					/* else fall through to Abs mode */
				}
				if (opmode[n] != LargeMode)
					break;
				if (! vm_is_u22bits(opvalue[n]))
					break;
				hexcode |= (((long)AbsMode) << 22);
				hexcode |= vm_u22bits(opvalue[n]);
				n++;
			}
			else if (ch == 'S') {
				if (opmode[n] != LargeMode)
					break;
				offset = opvalue[n] - p->addr;
				if (! vm_is_s22bits(offset))
					break;
				hexcode |= (((long)RelMode) << 22);
				hexcode |= vm_u22bits(offset);
				n++;
			}
			else if (ch == 'R') {
				if (opmode[n] != RegMode)
					break;
				if (! vm_is_u6bits(opvalue[n]))
					break;
				hexcode |= (((long)RegMode) << 22);
				hexcode |= vm_u6bits(opvalue[n]);
				n++;
			}
			else if (ch == '-') {
				/* nothing to do here; operand ignored */
			}
			else if (ch == 'L') {
				if (opmode[n] != LargeMode)
					break;
				if (! vm_is_s14bits(opvalue[n]))
					break;
				hexcode |= (((long)LargeMode) << 22);
				hexcode |= (vm_u14bits(opvalue[n]) << 8);
				n++;
			}
			else if (ch == 'l') {
				if (opmode[n] != LargeMode)
					break;
				if (! vm_is_s6bits(opvalue[n]))
					break;
				hexcode |= (((long)SmallMode) << (8*(2-i)+6));
				hexcode |= (vm_u6bits(opvalue[n]) << (8*(2-i)));
				n++;
			}
			else if (ch == 'r') {
				if (opmode[n] != RegMode)
					break;
				if (! vm_is_u6bits(opvalue[n]))
					break;
				hexcode |= (opmode[n] << (8*(2-i)+6));
				hexcode |= (vm_u6bits(opvalue[n]) << (8*(2-i)));
				n++;
			}
			else if ((ch == 'c') || (ch == 'i') || (ch == 'd')
				|| (ch == 's') || (ch == 'a'))
			{
				if (opmode[n] != CellMode)
					break;
				if (! vm_is_u6bits(opvalue[n]))
					break;
				hexcode |= (opmode[n] << (8*(2-i)+6));
				hexcode |= (vm_u6bits(opvalue[n]) << (8*(2-i)));
				n++;
			}
			else {
				error("A predefined operand set was %s: ", set->opchar);
				return 0;
			}
		}
	}
	gen_error(p, "Could not assemble instruction %ld: %ld", p->addr, opcode);
	return 0;
}

static long encode_op(Program p, Code code, long opcode,
	int numops, Operand * ops)
{
	int i, success;
	long hexcode = 0;
	long opmode[3];
	long opvalue[3];
	int relative_branch = 0;
	Token t;

	for (i=0; i < numops; i++) {
		t = ops[i].token;
		if (t) {
			opvalue[i] = t->num;
			switch (t->kind) {
			 case INTVAL:	opmode[i] = LargeMode; break;
			 case REG:	opmode[i] = RegMode; break;
			 case CELL:	opmode[i] = CellMode; break;
			 case IDENT:
				gen_error(p, "Label \"%s\" on line %ld used but never defined\n",
						t->name, t->line);
				opmode[i] = LargeMode;
				opvalue[i] = 0;
				break;
			 default:
				gen_error(p, "Invalid operand mode on line %ld\n",
						t->line);
				opmode[i] = LargeMode;
				opvalue[i] = 0;
			}
		}
		else if (ops[i].expr) {
			opmode[i] = LargeMode;
			opvalue[i] = eval_expr(p, ops[i].expr);
			if ((i==0) && (numops==1) && (code->op->num < 0x20))
				if (is_relative_branch(ops[i].expr)) {
					opvalue[i] += p->addr;
					relative_branch++;
				}
		}
	}
	success = encode(p, opcode, numops, opmode, opvalue, relative_branch,
			&hexcode);

	if (! success) {
		print_code(code);
		note("\n");
	}

	return hexcode;
}

/*
 *  Evaluate each element in the CODE section of the program
 *  by encoding it into hex.
 *  Return the size of the section.
 */
static long eval_code_section(Program p, CodeArray *array)
{
	long i;
	long size;
	Code code;

	size = array->size;

	for (i=0; i < size; i++) {
		code = array->data[i];
		code->hexcode = encode_op(p, code, code->op->num,
				code->noperands, code->operands);
	}
	return size;
}

/*
 *  Evaluate each element in the INT section of the program
 *  to check there are no labels used but not defined.
 *  Also save time by recording the result of that evaluation
 *  into a new array of integers, for later output.
 *  Return the size of the section.
 */
static long eval_int_section(Program p, ExprArray *array)
{
	long i;
	long size;
	Expr expr;
	long value;

	size = array->size;

	p->ints = new_intarray();

	for (i=0; i < size; i++) {
		expr = array->data[i];
		value = eval_expr(p, expr);
		intarray_append(p->ints, value);
	}
	return size;
}

/*
 *  Return the size of the DOUBLE section.
 */
static long eval_double_section(Program p, DoubleArray *array)
{
	return array->size;
}

/*
 *  Return the size of the STRING section.
 */
static long eval_string_section(Program p, StringArray *array)
{
	return array->size;
}

void check_program(Program p)
{
	p->Nint    = eval_int_section(p, p->exprs);
	p->Ndouble = eval_double_section(p, p->doubles);
	p->Nstring = eval_string_section(p, p->strings);
	p->Ncode   = eval_code_section(p, p->codes);
}

int gen_program(Program p, char *outname)
{
	check_program(p);
	if (p->error_count > 0)
		return 0;
	return 1;
}

/*
 *  Save each individual kind of data.
 */
static void save_code(FILE *file, Code code)
{
	fprintf(file, "%08lx\n", code->hexcode);
}

static void save_int(FILE *file, long value)
{
	fprintf(file, "%ld\n", value);
}

static void save_double(FILE *file, double value)
{
	fprintf(file, "%g\n", value);
}

static void save_string(FILE *file, char *s)
{
	long i, length;

	if (s == NULL) {
		/* this circumstance is possible if a STRING array is
		   larger than the number of elements assigned to it */
		fprintf(file, "\n");
		return;
	}

	length = strlen(s);

	for (i=1; i < length-1; i++)		// omit quotes around string
		fprintf(file, "%c", s[i]);
	fprintf(file, "\n");
}

/*
 *  Write entire sections of data to the output file.
 */
static void write_code_section(FILE *file, CodeArray *array)
{
	long i, size;

	size = array->size;
	for (i=0; i < size; i++)
		save_code(file, array->data[i]);
}

static void write_int_section(FILE *file, IntArray *array)
{
	long i, size;

	size = array->size;
	for (i=0; i < size; i++)
		save_int(file, array->data[i]);
}

static void write_double_section(FILE *file, DoubleArray *array)
{
	long i, size;

	size = array->size;
	for (i=0; i < size; i++)
		save_double(file, array->data[i]);
}

static void write_string_section(FILE *file, StringArray *array)
{
	long i, size;

	size = array->size;
	for (i=0; i < size; i++)
		save_string(file, array->data[i]);
}

void write_program(Program p)
{
	fprintf(p->file, "#!avm\n");
	fprintf(p->file, "%ld %ld %ld ", p->Nint, p->Ndouble, p->Nstring);
	fprintf(p->file, "%ld\n", p->Ncode);

	write_int_section(p->file, p->ints);
	write_double_section(p->file, p->doubles);
	write_string_section(p->file, p->strings);
	write_code_section(p->file, p->codes);
}

void output_program(Program p, FILE *outfile)
{
	p->file = outfile;
	write_program(p);
}

