/*
 *  Assembler Recursive Descent Parser:
 */

#include "asm.h"

/*
 *  Object allocation functions:
 */
#define alloc(obj) ((obj)zalloc(sizeof(struct obj)))

static Operand new_operand(void)
{
	Operand op;
	op.token = NULL;
	op.expr = NULL;
	return op;
}

static Code new_code(Token t)
{
	Code code = alloc(Code);
	memset(code, 0, sizeof(struct Code));
	code->op = t;
	return code;
}

static Expr new_expr(void)
{
	Expr e = alloc(Expr);
	memset(e, 0, sizeof(struct Expr));
	return e;
}

static void del_expr(Expr e)
{
	if (e) {
		if (e->left)
			del_expr(e->left);
		if (e->right)
			del_expr(e->right);
		zfree(e);
	}
}

static Program new_program(void)
{
	Program p = alloc(Program);
	memset(p, 0, sizeof(struct Program));

	p->codes   = new_codearray();
	p->exprs   = new_exprarray();
	p->ints    = NULL;		// NULL until evaluating the Exprs
	p->doubles = new_doublearray();
	p->strings = new_stringarray();
	return p;
}

/*
 *  Utilities to consume many tokens, or find a particular token.
 */
Token eat_all(Program p, int kind)
{
	Token t;

	t = peek_token(p);
	while (t->kind == kind) {
		t = get_token(p);
		t = peek_token(p);
		if (t->kind == EOF)
			break;
	}
	return t;
}

Token skip_to(Program p, int kind)
{
	Token t;

	t = peek_token(p);
	if ((t->kind != EOF) && (t->kind != kind))
		if (p->error_count == 0)
			parser_error(p, "Skipping invalid input");
	while ((t->kind != EOF) && (t->kind != kind)) {
		t = get_token(p);
		t = peek_token(p);
	}
	return t;
}

/*
 *  Type checking helper functions:
 */

int is_type(int kind)
{
	switch (kind) {
	  case INT: case DOUBLE: case STRING:
		return 1;
	  default:
		return 0;
	}
}

int is_unary_op(int kind)
{
	switch (kind) {
	  case BITWISE_NOT: case PLUS: case MINUS:
		return 1;
	  default:
		return 0;
	}
}

int is_relation_op(int kind)
{
	switch (kind) {
	  case LT: case LTEQ: case GT: case GTEQ:
		return 1;
	  default:
		return 0;
	}
}

int is_equal_op(int kind)
{
	switch (kind) {
	  case EQUALITY: case INEQUALITY:
		return 1;
	  default:
		return 0;
	}
}

int is_shift_op(int kind)
{
	switch (kind) {
	  case LSHIFT: case RSHIFT:
		return 1;
	  default:
		return 0;
	}
}

int is_add_op(int kind)
{
	switch (kind) {
	  case PLUS: case MINUS:
		return 1;
	  default:
		return 0;
	}
}

int is_mul_op(int kind)
{
	switch (kind) {
	  case STAR: case SLASH: case MOD:
		return 1;
	  default:
		return 0;
	}
}

int is_expr_start(int kind)
{
	switch (kind) {
	  case INTVAL: case LPARENTH: case IDENT:
	  case PLUS: case MINUS: case BITWISE_NOT:
		return 1;
	  default:
		return 0;
	}
}

int is_double_start(int kind)
{
	switch (kind) {
	  case PLUS: case MINUS: case INTVAL: case DOUBLEVAL:
		return 1;
	  default:
		return 0;
	}
}

/*
 *  Adding objects to the program:
 */

void add_operand(Program p, Code code, Operand op)
{
	if (code->noperands == 3)
		parser_error(p, "Instruction has more than three operands");
	else {
		code->operands[code->noperands] = op;
		code->noperands++;
	}
}


/*
 *  Determine what a label means:
 */
void add_label(Program p, Token t)
{
	if (t->kind == INTVAL) {
		error("label %s defined as %ld redefined (was %ld on line %ld)",
			t->name, t->num, p->addr, t->line);
		parser_error(p, "Label defined more than once");
	}
	else if (t->kind == IDENT) {
		t->kind = INTVAL;
		t->num = p->addr;
	}
}

/*
 *  Recursive descent parser:
 */
Expr parse_value(Program p)
{
	Token t;
	Expr e;

	t = peek_token(p);
	if ((t->kind == IDENT) || (t->kind == INTVAL)) {
		t = get_token(p);
		e = new_expr();
		e->token = t;
		e->kind = INTVAL;
	}
	else if (t->kind == LPARENTH) {
		e = parse_expr(p);
	}
	else {
		parser_error(p, "Expected integer value or label in expression");
		e = new_expr();
	}

	return e;
}

Expr parse_unary_expr(Program p)
{
	Token t;
	Expr e;

	t = peek_token(p);
	if (is_unary_op(t->kind)) {
		t = get_token(p);	// + - ~
		e = new_expr();
		e->right = parse_value(p);
		e->op = t->kind;
		e->kind = e->right->kind;
		if (e->kind != INTVAL)
			parser_error(p, "Unary operator cannot work on this type");
	}
	else {
		e = parse_value(p);
	}
	return e;
}

Expr parse_mul_expr(Program p)
{
	Token t;
	Expr e, e2;

	e = parse_unary_expr(p);
	t = peek_token(p);
	while (is_mul_op(t->kind)) {
		e2 = e;
		e = new_expr();
		e->left = e2;
		t = get_token(p);	// * / %
		e->op = t->kind;
		e->right = parse_unary_expr(p);
		if (e->right->kind != e->left->kind)
			parser_error(p, "Type mismatch in mulop expression");
		e->kind = e->left->kind;
		t = peek_token(p);
	}
	return e;
}

Expr parse_add_expr(Program p)
{
	Token t;
	Expr e, e2;

	e = parse_mul_expr(p);
	t = peek_token(p);
	while (is_add_op(t->kind)) {
		e2 = e;
		e = new_expr();
		e->left = e2;
		t = get_token(p);	// + -
		e->op = t->kind;
		e->right = parse_mul_expr(p);
		if (e->right->kind != e->left->kind)
			parser_error(p, "Type mismatch in addop expression");
		e->kind = e->left->kind;
		t = peek_token(p);
	}
	return e;
}

Expr parse_shift_expr(Program p)
{
	Token t;
	Expr e, e2;

	e = parse_add_expr(p);
	t = peek_token(p);
	while (is_shift_op(t->kind)) {
		e2 = e;
		e = new_expr();
		e->left = e2;
		t = get_token(p);	// << >>
		e->op = t->kind;
		e->right = parse_add_expr(p);
		if (e->right->kind != INTVAL)
			parser_error(p, "Right hand side of shiftop must be int");
		e->kind = e->left->kind;
		t = peek_token(p);
	}
	return e;
}

#if 0

Expr parse_relation_expr(Program p)
{
	Token t;
	Expr e, e2;

	e = parse_shift_expr(p);
	t = peek_token(p);
	if (is_relation_op(t->kind)) {
		e2 = e;
		e = new_expr();
		e->left = e2;
		t = get_token(p);	// < <= > >=
		e->op = t->kind;
		e->right = parse_shift_expr(p);
		if (e->right->kind != e->left->kind)
			parser_error(p, "Type mismatch in relational expression");
		e->kind = BOOLEAN;
	}
	return e;
}

Expr parse_equals_expr(Program p)
{
	Token t;
	Expr e, e2;

	e = parse_relation_expr(p);
	t = peek_token(p);
	if (is_equal_op(t->kind)) {
		e2 = e;
		e = new_expr();
		e->left = e2;
		t = get_token(p);	// == !=
		e->op = t->kind;
		e->right = parse_relation_expr(p);
		if (e->right->kind != e->left->kind)
			parser_error(p, "Type mismatch in equality expression");
		e->kind = BOOLEAN;
	}
	return e;
}

#endif

Expr parse_bitwise_and(Program p)
{
	Token t;
	Expr e, e2;

	e = parse_shift_expr(p);	// was EqualsExpr
	t = peek_token(p);
	while (t->kind == BITWISE_AND) {
		e2 = e;
		e = new_expr();
		e->left = e2;
		t = get_token(p);	// &
		e->op = t->kind;
		e->right = parse_shift_expr(p);	// was EqualsExpr
		if (e->right->kind != e->left->kind)
			parser_error(p, "Type mismatch in bitwise AND expression");
		e->kind = e->left->kind;
		t = peek_token(p);
	}
	return e;
}

Expr parse_bitwise_xor(Program p)
{
	Token t;
	Expr e, e2;

	e = parse_bitwise_and(p);
	t = peek_token(p);
	while (t->kind == BITWISE_XOR) {
		e2 = e;
		e = new_expr();
		e->left = e2;
		t = get_token(p);	// ^
		e->op = t->kind;
		e->right = parse_bitwise_and(p);
		if (e->right->kind != e->left->kind)
			parser_error(p, "Type mismatch in bitwise XOR expression");
		e->kind = e->left->kind;
		t = peek_token(p);
	}
	return e;
}

Expr parse_bitwise_or(Program p)
{
	Token t;
	Expr e, e2;

	e = parse_bitwise_xor(p);
	t = peek_token(p);
	while (t->kind == BITWISE_OR) {
		e2 = e;
		e = new_expr();
		e->left = e2;
		t = get_token(p);	// |
		e->op = t->kind;
		e->right = parse_bitwise_xor(p);
		if (e->right->kind != e->left->kind)
			parser_error(p, "Type mismatch in bitwise OR expression");
		e->kind = e->left->kind;
		t = peek_token(p);
	}
	return e;
}

#if 0

Expr parse_conditional_and(Program p)
{
	Token t;
	Expr e, e2;

	e = parse_bitwise_or(p);
	t = peek_token(p);
	while (t->kind == AND) {
		e2 = e;
		e = new_expr();
		e->left = e2;
		t = get_token(p);	// boolean and
		e->op = t->kind;
		e->right = parse_bitwise_or(p);
		if (e->right->kind != e->left->kind)
			parser_error(p, "Type mismatch in boolean AND expression");
		if (e->right->kind != BOOLEAN)
			parser_error(p, "Expression after boolean AND not boolean?");
		e->kind = e->left->kind;
		t = peek_token(p);
	}
	return e;
}

Expr parse_conditional_or(Program p)
{
	Token t;
	Expr e, e2;

	e = parse_conditional_and(p);
	t = peek_token(p);
	while (t->kind == OR) {
		e2 = e;
		e = new_expr();
		e->left = e2;
		t = get_token(p);	// boolean or
		e->op = t->kind;
		e->right = parse_conditional_and(p);
		if (e->right->kind != e->left->kind)
			parser_error(p, "Type mismatch in boolean OR expression");
		if (e->right->kind != BOOLEAN)
			parser_error(p, "Expression after boolean OR not boolean?");
		e->kind = e->left->kind;
		t = peek_token(p);
	}
	return e;
}

#endif

Expr parse_expr(Program p)
{
	Expr e;

	e = parse_bitwise_or(p);	// was CondOR
	return e;
}

Operand parse_operand(Program p)
{
	Token t;
	Operand op;
	Expr e;

	op = new_operand();
	t = peek_token(p);
	switch (t->kind) {
	 case REG:
	 case CELL:
		op.token = get_token(p);
		break;
	 case IDENT:
	 case INTVAL:
	 case LPARENTH:
	 case PLUS:
	 case MINUS:
	 case BITWISE_NOT:
		e = parse_expr(p);
		op.expr = e;
		break;
	 default:
		parser_error(p, "Invalid operand");
	}
	return op;
}

int print_code_definition(Program p)
{
	Code code;
	Token t = peek_token(p);

	switch (t->kind) {
	 case NEWLINE:
		t = get_token(p);
		return 1;
	 case OPCODE:
		code = new_code(get_token(p));
		codearray_append(p->codes, code);
		p->addr++;
		t = peek_token(p);
		if ((t->kind != NEWLINE) && (t->kind != EOF)) {
			add_operand(p, code, parse_operand(p));
			t = peek_token(p);
			if ((t->kind != COMMA) && (t->kind != NEWLINE)) {
				parser_error(p, "Expected newline or comma after operand");
				skip_to(p, NEWLINE);
			}
		}
		t = peek_token(p);
		if (t->kind == COMMA) {
			t = get_token(p);
			add_operand(p, code, parse_operand(p));
			t = peek_token(p);
			if ((t->kind != COMMA) && (t->kind != NEWLINE)) {
				parser_error(p, "Expected newline or comma after operand");
				skip_to(p, NEWLINE);
			}
		}
		t = peek_token(p);
		if (t->kind == COMMA) {
			t = get_token(p);
			add_operand(p, code, parse_operand(p));
			t = peek_token(p);
			if (t->kind != NEWLINE) {
				parser_error(p, "Expected newline after operand");
				skip_to(p, NEWLINE);
			}
		}
		return eat_token(p, NEWLINE, "Expected newline after instruction");
	 case IDENT:
		add_label(p, get_token(p));
		return eat_token(p, COLON, "Expected colon after label");
	}
	return 0;
}

void parse_int_array(Program p, Token t)
{
	ExprArray *array;
	Expr e = NULL;
	long i, max;

	e = parse_expr(p);
	eat_token(p, RBRACKET, "Expected ] after array size");
	max = eval_expr(p, e);
	del_expr(e);

	array = new_exprarray();
	i = 0;
	t = peek_token(p);
	while (is_expr_start(t->kind)) {
		if (++i > max)
			parser_error(p, "More integers in array than expected");
		e = parse_expr(p);
		exprarray_append(array, e);
		t = get_token(p);
		if (t->kind == NEWLINE)
			break;
		else if (t->kind == EOF)
			break;
		else if (t->kind == COMMA)
			t = peek_token(p);
		else {
			parser_error(p, "Expected newline or comma");
			break;
		}
	}
	for (i=0; i < array->size; i++)
		exprarray_append(p->exprs, array->data[i]);
	for (; i < max; i++)
		exprarray_append(p->exprs, NULL);
	del_exprarray(array);
	p->addr += max;
}

int print_int_definition(Program p)
{
	Expr e;
	Token t = peek_token(p);

	switch (t->kind) {
	 case NEWLINE:
		t = get_token(p);
		return 1;
	 case INTVAL:
	 case LPARENTH:
	 case PLUS:
	 case MINUS:
	 case BITWISE_NOT:
		e = parse_expr(p);
		exprarray_append(p->exprs, e);
		p->addr++;
		return 1;
		t = peek_token(p);
		if (t->kind == EOF)
			return 1;
		return eat_token(p, NEWLINE, "Expected newline after integer constant");
	 case IDENT:
		add_label(p, get_token(p));
		return eat_token(p, COLON, "Expected colon after label");
	 case LBRACKET:
		parse_int_array(p, get_token(p));
		return 1; 
	}
	return 0;
}

void parse_double_array(Program p, Token t)
{
	DoubleArray *array;
	Expr e = NULL;
	long i, max;
	int sign;
	double d;

	e = parse_expr(p);
	eat_token(p, RBRACKET, "Expected ] after array size");
	max = eval_expr(p, e);
	array = new_doublearray();
	del_expr(e);

	i = 0;
	t = peek_token(p);
	while (is_double_start(t->kind)) {
		if (++i > max)
			parser_error(p, "More numbers in array than expected");

		sign = 0;
		if (t->kind == PLUS) {
			t = get_token(p);
			sign = +1;
			t = peek_token(p);
		}
		else if (t->kind == MINUS) {
			t = get_token(p);
			sign = -1;
			t = peek_token(p);
		}
		if ((sign) && ((t->kind != INTVAL) && (t->kind != DOUBLEVAL)) )
		{
			parser_error(p, "Expected floating-point number after sign");
			break;
		}

		t = get_token(p);
		if (t->kind == INTVAL) {
			if (sign == -1)
				d = 0.0 - t->num;
			else
				d = 0.0 + t->num;
		}
		else if (t->kind == DOUBLEVAL) {
			if (sign == -1)
				d = 0.0 - t->real;
			else
				d = t->real;
		}
		else {
			d = 0.0;	// should never reach here
		}
		doublearray_append(array, d);
		t = get_token(p);
		if (t->kind == NEWLINE)
			break;
		else if (t->kind == EOF)
			break;
		else if (t->kind == COMMA)
			t = peek_token(p);
		else {
			parser_error(p, "Expected newline or comma");
			break;
		}
	}
	for (i=0; i < array->size; i++)
		doublearray_append(p->doubles, array->data[i]);
	for (; i < max; i++)
		doublearray_append(p->doubles, 0.0);
	del_doublearray(array);
	p->addr += max;
}

int print_double_definition(Program p)
{
	Token t = peek_token(p);
	int sign = 0;

	if (t->kind == PLUS) {
		t = get_token(p);
		sign = +1;
		t = peek_token(p);
	}
	else if (t->kind == MINUS) {
		t = get_token(p);
		sign = -1;
		t = peek_token(p);
	}
	if ((sign) && ((t->kind != INTVAL) && (t->kind != DOUBLEVAL)) ) {
		parser_error(p, "Expected floating-point number after sign");
		return 0;
	}

	switch (t->kind) {
	 case NEWLINE:
		t = get_token(p);
		return 1;
	 case INTVAL:
		t = get_token(p);
		if (sign == -1)
			doublearray_append(p->doubles, 0.0 - t->num);
		else
			doublearray_append(p->doubles, 0.0 + t->num);
		p->addr++;
		return 1;
	 case DOUBLEVAL:
		t = get_token(p);
		if (sign == -1)
			doublearray_append(p->doubles, 0.0 - t->real);
		else
			doublearray_append(p->doubles, t->real);
		p->addr++;
		return 1;
		t = peek_token(p);
		if (t->kind == EOF)
			return 1;
		return eat_token(p, NEWLINE, "Expected newline after floating-point number");
	 case IDENT:
		add_label(p, get_token(p));
		return eat_token(p, COLON, "Expected colon after label");
	 case LBRACKET:
		parse_double_array(p, get_token(p));
		return 1; 
	}
	return 0;
}

void parse_string_array(Program p, Token t)
{
	StringArray *array;
	Expr e = NULL;
	long i, max;

	e = parse_expr(p);
	eat_token(p, RBRACKET, "Expected ] after array size");
	max = eval_expr(p, e);
	del_expr(e);

	array = new_stringarray();
	i = 0;
	t = peek_token(p);
	while (t->kind == STRINGVAL) {
		if (++i > max)
			parser_error(p, "More strings in array than expected");
		t = get_token(p);
		stringarray_append(array, t->name);
		t = get_token(p);
		if (t->kind == NEWLINE)
			break;
		else if (t->kind == EOF)
			break;
		else if (t->kind == COMMA)
			t = peek_token(p);
		else {
			parser_error(p, "Expected newline or comma");
			break;
		}
	}
	for (i=0; i < array->size; i++)
		stringarray_append(p->strings, array->data[i]);
	for (; i < max; i++)
		stringarray_append(p->strings, NULL);
	del_stringarray(array);
	p->addr += max;
}

int print_string_definition(Program p)
{
	Token t = peek_token(p);

	switch (t->kind) {
	 case NEWLINE:
		t = get_token(p);
		return 1;
	 case STRINGVAL:
		t = get_token(p);
		stringarray_append(p->strings, t->name);
		p->addr++;
		return 1;
		t = peek_token(p);
		if (t->kind == EOF)
			return 1;
		return eat_token(p, NEWLINE, "Expected newline after string");
	 case IDENT:
		add_label(p, get_token(p));
		return eat_token(p, COLON, "Expected colon after label");
	 case LBRACKET:
		parse_string_array(p, get_token(p));
		return 1; 
	}
	return 0;
}

void parse_code_section(Program p)
{
	Token t;

	eat_all(p, NEWLINE);
	t = peek_token(p);
	if (t->kind == CODE) {
		t = get_token(p);
		while (print_code_definition(p))
			continue;
	}
}

void parse_int_section(Program p)
{
	Token t;

	eat_all(p, NEWLINE);
	t = peek_token(p);
	if (t->kind == INT) {
		t = get_token(p);
		while (print_int_definition(p))
			continue;
	}
}

void parse_double_section(Program p)
{
	Token t;

	eat_all(p, NEWLINE);
	t = peek_token(p);
	if (t->kind == DOUBLE) {
		t = get_token(p);
		while (print_double_definition(p))
			continue;
	}
}

void parse_string_section(Program p)
{
	Token t;

	eat_all(p, NEWLINE);
	t = peek_token(p);
	if (t->kind == STRING) {
		t = get_token(p);
		while (print_string_definition(p))
			continue;
	}
}

void parse_sections(Program p)
{
	Token t;

	p->addr = 0;		/* set data address to zero */
	parse_int_section(p);
	parse_double_section(p);
	parse_string_section(p);

	p->addr = 0;		/* set instruction address to zero */
	parse_code_section(p);

	eat_all(p, NEWLINE);
	t = peek_token(p);
	if (t->kind != EOF)
		parser_error(p, "Unassembled code");
}

Program parse_program(FILE *f)
{
	Program p;

	init_tokeniser();
	p = new_program();
	p->file = f;

	/* test tokeniser */
/*
	for (t = get_token(p); t; t = get_token(p)) {
		if (t->kind == EOF)
			break;
		if (t->kind == ERROR) {
			error("tokeniser error");
			break;
		}
	}
*/

	parse_sections(p);

	return p;
}

