/*
 *  Debug Printing:
 */

#include "asm.h"

static void print_int(long i)
{
	note("%ld", i);
}

static void print_double(double d)
{
	note("%g", d);
}

static void print_string(char *s)
{
	long i, length;

	if (s == NULL) {
		/* could happen if there are more array elements than given */
		return;
	}

	length = strlen(s);
	for (i=1; i<length-1; i++)	// omit quotes around string
		note("%c", s[i]);

	//note("%s", s);
}

static void print_int_token(Token t)
{
	print_int(t->num);
}

static void print_double_token(Token t)
{
	print_double(t->real);
}

static void print_string_token(Token t)
{
	note("%s", t->name);
}

static void print_expr(Expr e)
{
	if (e == NULL)
		return;
	if (e->token) {
		switch (e->token->kind) {
		 case IDENT:  print_string_token(e->token); break;
		 case INTVAL: print_int_token(e->token); break;
		 case DOUBLEVAL: print_double_token(e->token); break;
		 default:
			print_token(e->token);
			error("type in expression either not constant or not integer (%d)", e->token->kind);
		}
	}
	else {
		if (e->left && e->right) {
			note("(");
			print_expr(e->left);
			note(" %s ", TokenNames[e->op]);
			print_expr(e->right);
			note(")");
		}
		else if (e->left)
			print_expr(e->left);
		else if (e->right) {
			note(" %s ", TokenNames[e->op]);
			print_expr(e->right);
		}
	}
}

static void print_operand(Operand op)
{
	if (op.token)
		note("%s", op.token->name);
	else if (op.expr)
		print_expr(op.expr);
}

void print_code(Code code)
{
	int i;

	note("%s", code->op->name);

	for (i=0; i < code->noperands; i++) {
		if (i == 0)
			note("\t");
		else
			note(", ");
		print_operand(code->operands[i]);
	}
}

void print_code_section(char *section, CodeArray *array)
{
	long i, size;

	note("%s\n\n", section);
	size = array->size;
	for (i=0; i < size; i++) {
		print_code(array->data[i]);
		note("\n");
	}
	note("\n");
}

void print_int_section(char *section, ExprArray *array)
{
	long i, size;

	note("%s\n\n", section);
	size = array->size;
	for (i=0; i < size; i++) {
		if (array->data[i] == NULL)
			note("0");
		else
			print_expr(array->data[i]);
		note("\n");
	}
	note("\n");
}

void print_double_section(char *section, DoubleArray *array)
{
	long i, size;

	note("%s\n\n", section);
	size = array->size;
	for (i=0; i < size; i++) {
		print_double(array->data[i]);
		note("\n");
	}
	note("\n");
}

void print_string_section(char *section, StringArray *array)
{
	long i, size;

	note("%s\n\n", section);
	size = array->size;
	for (i=0; i < size; i++) {
		print_string(array->data[i]);
		note("\n");
	}
	note("\n");
}

void print_program(Program p)
{
	print_int_section(TokenNames[INT], p->exprs);
	print_double_section(TokenNames[DOUBLE], p->doubles);
	print_string_section(TokenNames[STRING], p->strings);
	print_code_section(TokenNames[CODE], p->codes);
}

