/*
 *  Assembler Tokeniser:
 */

#include "asm.h"

/*
 *  Global input buffer which is automatically resized as necessary.
 */
  static char * buffer = NULL;
  static long buf_size = 0;
  static long buf_pos = 0;

/*
 *  Input file line number, starts from 1 each time.
 */
  static long line_number = 1;

/*
 *  Allow buffering of Tokens for lookahead:
 */

  static Token current_token = NULL;
  static Token cached_token = NULL;

/*
 *  Table of keywords
 */
  char * TokenNames[] = {
	"#ERROR#", "#EOF#", "#\\n#",
	"CODE", "INT", "DOUBLE", "STRING",
	"#OPCODE#", "#INT#", "#DBL#", "#STR#",
	"#IDENT#", "#R#", "#[R]#",
	"<<",	">>",	"(",	")",
	"{",	"}",	"[",	"]",
	"==",	"!=",	"<=",	">=",	"<",	">",	"=",
	",",	";",	"*",	"/",	"%",	"+",	"-",
	"&",	"|",	"^",	"~",
	"\\",	"\"",	".",	":",
	"pc",
  };

/*
 *  Symbol Table (for storing all Tokens)
 *
 *  All keywords, punctuation and identifiers are placed
 *  into the symbol table. Pointers to Tokens are returned
 *  by the tokeniser, but the memory is entirely managed
 *  within the symbol table, so the rest of the program
 *  need never free memory. Tokens can simply be discarded.
 *
 *  One function manages the symbol table memory: init_tokeniser
 *  must be called every time a new file is being parsed.
 */

enum {
	NHASH=1024,
	MAXTOKEN = 1024,
};


static int hash_table_used = 0;
static Token hash_table[NHASH];

Token new_token(void)
{
	Token t;

	t = (Token) zalloc(sizeof(struct Token));
	t->next = NULL;
	t->line = line_number;
	t->kind = ERROR;
	t->num = 0;
	t->real = 0.0;
	t->name = NULL;
	return t;
}

static void delToken(Token t)
{
	zfree(t->name);
	zfree(t);
}

static int hash(char *name)
{
	unsigned long h;

	for(h=0; *name; name++)
		h = h*31 + *name;
	return h%NHASH;
}

/*
 *  Set the hash table to all NULL.
 */
static void init_hash_table(void)
{
	int i;

	for (i=0; i < NHASH; i++)
		hash_table[i] = NULL;
}

/*
 *  Hash lookup
 */
static Token find_token(char *name)
{
	Token t;

	for (t = hash_table[hash(name)]; t; t=t->next)
		if (!strcmp(name, t->name))
			return t;
	return NULL;
}

/*
 *  Hash insert (ensures uniqueness)
 */
Token add_token(Token t)
{
	int h;
	Token prev;

	prev = find_token(t->name);
	if (prev) {
		if (prev != t)
			delToken(t);
		return prev;
	}
	else {
		h = hash(t->name);
		t->next = hash_table[h];
		hash_table[h] = t;
		return t;
	}
}

/*
 *  Debug printing:
 */

void print_token(Token t)
{
	char *name;

	fprintf(stderr, "token: ");
	if (t->kind == EOF)
		name = TokenNames[ENDOFFILE];
	else
		name = TokenNames[t->kind];
	fprintf(stderr, "line=%ld\t", t->line);
	fprintf(stderr, "kind=%2d %s\t", t->kind, name);
	fprintf(stderr, "int=%ld\t", t->num);
	fprintf(stderr, "real=%f\t", t->real);
	fprintf(stderr, "str=\"%s\"", t->name ? t->name : "");
	fprintf(stderr, "\n");
}

void print_all_tokens(void)
{
	int i;
	Token t;

	for (i=0; i<NHASH; i++) {
		fprintf(stderr, "hash value %d:\n", i);
		for (t = hash_table[i]; t; t=t->next)
			print_token(t);
	}
}

/*
 *  Remove all identifiers from the hash table, but leave keywords.
 */
static int is_keyword(Token t)
{
	return (t->line > 0);
}

static void clear_identifiers(void)
{
	int i;
	Token t, next;

	for (i=0; i < NHASH; i++) {
		t = hash_table[i];
		while (t && is_keyword(t)) {
			hash_table[i] = t->next;
			next = t->next;
			delToken(t);
			t = next;
		}
	}
}

/*
 *  Add all keywords to the hash table.
 */
static void add_keywords(void)
{
	int i;
	Token t;
	char buf[30];

	/* add keywords */
	for (i=0; i < (sizeof(TokenNames)/sizeof(TokenNames[0])); i++) {
		if (TokenNames[i] == NULL)
			continue;
		t = new_token();
		t->line = 0; /* keyword */
		t->kind = i;
		if (i == ENDOFFILE)
			t->kind = EOF;
		t->name = zstrdup(TokenNames[i]);
		add_token(t);
	}
	/* add instruction names */
	for (i=0; i < (sizeof(vm_opnames)/sizeof(vm_opnames[0])); i++) {
		if (vm_opnames[i] == NULL)
			continue;
		t = new_token();
		t->line = 0; /* instruction name */
		t->kind = OPCODE;
		t->num  = i; /* opcode */
		t->name = zstrdup(vm_opnames[i]);
		add_token(t);
	}
	/* add register identifiers r0 ... r63 */
	for (i=0; i < Nregs; i++) {
		sprintf(buf, "r%i", i);
		t = new_token();
		t->line = 0;
		t->kind = REG;
		t->num  = i; /* register operand code */
		t->name = zstrdup(buf);
		add_token(t);
	}
	/* add register identifiers r00 ... r09 */
	for (i=0; (i < 10) && (i < Nregs); i++) {
		sprintf(buf, "r0%i", i);
		t = new_token();
		t->line = 0;
		t->kind = REG;
		t->num  = i; /* register operand code */
		t->name = zstrdup(buf);
		add_token(t);
	}
	/* add cell identifiers [r0 ... r63] */
	for (i=0; i < Nregs; i++) {
		sprintf(buf, "[r%i]", i);
		t = new_token();
		t->line = 0;
		t->kind = CELL;
		t->num  = i; /* cell operand code */
		t->name = zstrdup(buf);
		add_token(t);
	}
	/* add cell identifiers [r00 ... r09] */
	for (i=0; (i < 10) && (i < Nregs); i++) {
		sprintf(buf, "[r0%i]", i);
		t = new_token();
		t->line = 0;
		t->kind = CELL;
		t->num  = i; /* cell operand code */
		t->name = zstrdup(buf);
		add_token(t);
	}
	/* add sp (make it equal to r63) */
	{
		sprintf(buf, "sp");
		t = new_token();
		t->line = 0;
		t->kind = REG;
		t->num  = SP; /* register operand code */
		t->name = zstrdup(buf);
		add_token(t);
	}
	/* add [sp] (make it equal to [r63]) */
	{
		sprintf(buf, "[sp]");
		t = new_token();
		t->line = 0;
		t->kind = CELL;
		t->num  = SP; /* cell operand code */
		t->name = zstrdup(buf);
		add_token(t);
	}
	/* add fp (make it equal to r62) */
	{
		sprintf(buf, "fp");
		t = new_token();
		t->line = 0;
		t->kind = REG;
		t->num  = FP; /* register operand code */
		t->name = zstrdup(buf);
		add_token(t);
	}
	/* add [fp] (make it equal to [r62]) */
	{
		sprintf(buf, "[fp]");
		t = new_token();
		t->line = 0;
		t->kind = CELL;
		t->num  = FP; /* cell operand code */
		t->name = zstrdup(buf);
		add_token(t);
	}
	//print_all_tokens();
}

/*
 *  Error handler: prints the error + line number and stops parsing.
 */
void parser_error(Program p, char *msg)
{
	if (++(p->error_count) > MAX_ERRORS)
		return;
	if (msg)
		fprintf(stderr, "error: %s near line %ld\n", msg, line_number);
/*
	if (current_token) {
		fprintf(stderr, "this ");
		print_token(current_token);
	}
	if (cached_token) {
		fprintf(stderr, "next ");
		print_token(cached_token);
	}
*/
}

/*
 *  Low-level file input (reads binary or text, Dos, Unix or Mac):
 */
static int get_char(Program p)
{
	int ch;

	if (p->error_count > MAX_ERRORS)
		return EOF;
	ch = getc(p->file);
	if (ch == '\r') {
		ch = getc(p->file);
		if (ch != '\n')
			ungetc(ch, p->file);
		ch = '\n';
	}
	if (ch == '\n')
		line_number++;
	return ch;
}

static int unget_char(int ch, Program p)
{
	int value = ch;
	if (ch != EOF)
		value = ungetc(ch, p->file);
	if (ch == '\n')
		line_number--;
	return value;
}

static int peek_char(Program p)
{
	int ch = get_char(p);
	unget_char(ch, p);
	return ch;
}

/*
 *  Handle string-buffering:
 */
static void add_char(int ch)
{
	char *ptr;

	if (buf_pos == buf_size) {
		ptr = zrealloc(buffer, (buf_size * 2 + 1) * sizeof(char));
		if (ptr) {
			buffer = ptr;
			buf_size = buf_size * 2 + 1;
			buffer[buf_pos++] = ch;
		}
		else {
			fprintf(stderr, "ran out of memory\n");
			exit(1);
		}
	}
	else {
		buffer[buf_pos++] = ch;
	}
}

/*
 *  Handle spacing:
 *  Note: newlines are not counted as spaces by this tokeniser;
 *  rather, they are tokens in their own right.
 */
static void skip_spaces(Program p)
{
	int ch;

	while ((ch=get_char(p)) != EOF) {
		if ((ch == '\n') || (! isspace(ch))) {
			unget_char(ch, p);
			break;
		}
	}
}

/*
 *  Handle # comments:
 *  Assume we are about to read the start of a comment.
 */
static void skip_comment(Program p)
{
	int ch = get_char(p);

	if (ch == '#') {	// skip until end of line
		while ((ch = get_char(p)) != EOF) {
			if (ch == '\n') {
				unget_char(ch, p);
				break;
			}
		}
	}
	else {
		unget_char(ch, p);
	}
}

static int ishexdigit(int ch)
{
	if (isdigit(ch))    return 1;
	ch = tolower(ch);
	if (ch == 'a')      return 1;
	else if (ch == 'b') return 1;
	else if (ch == 'c') return 1;
	else if (ch == 'd') return 1;
	else if (ch == 'e') return 1;
	else if (ch == 'f') return 1;
	return 0;
}

static Token get_number_token(Program p)
{
	Token t;
	int ch;
	int base=10;

	// read in a plain number
	t = new_token();

	while ((ch = peek_char(p)) != EOF) {
		if ((buf_pos == 0) && (ch == '0')) {
			t->kind = INTVAL;
			add_char(get_char(p));
			ch = peek_char(p);
			if ((ch == 'x') || (ch == 'X')) {
				add_char(get_char(p));
				base=16;
			}
		}
		else if ((base == 16) && (ishexdigit(ch))) {
			t->kind = INTVAL;
			add_char(get_char(p));
		}
		else if ((base == 16) && (ch == '.')) {
			break;	/* found 0x1f7. */
		}
		else if (isdigit(ch)) {
			t->kind = INTVAL;
			add_char(get_char(p));
		}
		else if (ch == '.') {
			t->kind = DOUBLEVAL;
			add_char(get_char(p));
			if (! isdigit(ch = peek_char(p)))
				break; /* found 123.x */
			while (isdigit(ch = peek_char(p)))
				add_char(get_char(p));
			if (toupper(ch = peek_char(p)) == 'E') {
				add_char(get_char(p));
				ch = peek_char(p);
				if ((ch == '+') || (ch == '-'))
					add_char(get_char(p));
				while (isdigit(ch = peek_char(p)))
					add_char(get_char(p));
			}
			break;
		}
		else {
			break;
		}
	}
	add_char('\0');
	sscanf(buffer, "%li", &t->num);
	sscanf(buffer, "%lf", &t->real);

	return t;
}

/*
 *  Read an identifier: an instruction, keyword, label,
 *  register identifier (r0 ... r63) etc
 */
static Token get_identifier_token(Program p)
{
	Token t;
	int ch;

	t = new_token();
	t->kind = IDENT;

	// read in an identifier
	while ((ch = get_char(p)) != EOF) {
		if (isalpha(ch) || (ch == '_') || isdigit(ch))
			add_char(ch);
		else {
			unget_char(ch, p);
			break;
		}
	}
	add_char('\0');

	return t;
}

/*
 *  Read in punctuation marks, including memory cell identifiers
 *  such as [r7], [r4], [sp] etc. '[' followed by non-alphabetics is
 *  returned as separate tokens '[' then  whatever then ']' etc.
 */
static Token get_punct_token(Program p)
{
	Token t;
	int ch;

	t = new_token();

	// read in a punctuation mark
	if ((ch = get_char(p)) != EOF) {
		if (ispunct(ch) && (ch != '_')) {
			add_char(ch);

			switch (ch) {
			 case '[':
				ch = peek_char(p);
				if (! isalpha(ch))
					break;
				while (ch != ']') {
					ch = peek_char(p);
					if ((! isalpha(ch)) && (! isdigit(ch)))
						break;
					add_char(get_char(p));
				}
				if (ch == ']')
					add_char(get_char(p));
				break;

			 case '<':
				ch = get_char(p);
				if ((ch == '<') || (ch == '='))
					add_char(ch);
				else
					unget_char(ch, p);
				break;

			 case '>':
				ch = get_char(p);
				if ((ch == '>') || (ch == '='))
					add_char(ch);
				else
					unget_char(ch, p);
				break;

			 case '=': case '!':
				ch = get_char(p);
				if (ch == '=')
					add_char(ch);
				else
					unget_char(ch, p);
				break;

			 default:
				break;
			}
		}
		else {
			unget_char(ch, p);
		}
	}
	add_char('\0');

	return t;
}

static Token get_string_token(Program p)
{
	Token t;
	int ch, prev;

	t = new_token();
	t->kind = STRINGVAL;

	// read in the string until we hit a quote char
	ch = get_char(p);
	if (ch != '\"') {
		t->kind = ERROR;
		unget_char(ch, p);	// error
		return t;
	}

	add_char(ch);	// keep quote char

	prev = '\0';
	while ((ch = get_char(p)) != EOF) {
		if (ch == '\"') {
			if (prev == '\\') {
				buffer[buf_pos-1] = ch;
				// buf_pos guaranteed to be > 0
				// since prev char exists
			}
			else {
				add_char(ch);	// keep quote char
				break;
			}
		} else if (ch == '\\') {
			add_char(ch);
			if (prev == '\\')
				ch = ' ';	/* avoid carry-over */
		} else if (ch == '\t') {
			add_char('\\');
			add_char('t');
		} else if (ch == '\n') {
			parser_error(p, "newline within string");
		} else {
			add_char(ch);
		}
		prev = ch;
	}
	add_char('\0');

	return t;
}

static Token fetch_token(Program p)	/* fetch token from file */
{
	int ch;
	Token t = NULL;

	buf_pos = 0;

	// find next token
	while ((ch=get_char(p)) != EOF) {
		// find token
		if (ch == '\n') {
			t = find_token(TokenNames[NEWLINE]);
			break;
		}
		else if (isspace(ch)) {
			unget_char(ch, p);
			skip_spaces(p);
			continue;	// loop
		}
		else if (ch == '#') {
			unget_char(ch, p);
			skip_comment(p);
			continue;	// loop
		}
		else if (isdigit(ch)) {
			unget_char(ch, p);
			t = get_number_token(p);
			t->name = zstrdup(buffer);
		}
		else if (isalpha(ch)) {
			unget_char(ch, p);
			t = get_identifier_token(p);
			t->name = zstrdup(buffer);
		}
		else if (ch == '\"') {
			unget_char(ch, p);
			t = get_string_token(p);
			t->name = zstrdup(buffer);
		}
		else if (ispunct(ch)) {
			unget_char(ch, p);
			t = get_punct_token(p);
			t->name = zstrdup(buffer);
		}
		else {
			t = find_token(TokenNames[ERROR]);
			break;
		}
		// always break, except for comments and spaces
		break;
	}
	if ((ch == EOF) && (t == NULL)) {
		t = find_token(TokenNames[ENDOFFILE]);
	}
	t = add_token(t);
	//print_token(t);
	return t;
}

/*
 *  Initialise (or re-initialise) the tokeniser hash table.
 */
void init_tokeniser(void)
{
	if (hash_table_used)
		clear_identifiers();
	else {
		init_hash_table();
		add_keywords();
		hash_table_used = 1;
	}
	line_number = 1;
	current_token = NULL;
	cached_token = NULL;
}

Token get_token(Program p)	/* fetch next token from the file */
{
	current_token = cached_token;
	cached_token = NULL;
	if (! current_token)
		current_token = fetch_token(p);
	if (current_token->kind == ERROR)
		parser_error(p, "some weird error");
	return current_token;
}

Token peek_token(Program p)	/* peek ahead at next token */
{
	if (! cached_token)
		cached_token = fetch_token(p);
	return cached_token;
}

int eat_token(Program p, int kind, char *msg)
{
	Token t = get_token(p);
	if (t->kind != kind) {
		parser_error(p, msg);
		return 0;
	}
	return 1;
}

